/*
   Copyright 2005 Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in
      the documentation and/or other materials provided with the
      distribution.
    * Neither the name of Red Hat, Inc., nor the names of its
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
   IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
   PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
   OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "../config.h"
#include <sys/types.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/* Write to a file, handling transient errors. */
ssize_t
retry_write(int fd, unsigned char *buf, size_t length)
{
	size_t start;
	ssize_t i;
	start = 0;
	while (start < length) {
		i = write(fd, buf + start, length - start);
		switch (i) {
		case 0:
			return start;
			break;
		case -1:
			switch (errno) {
			case EAGAIN:
				continue;
			default:
				return -1;
			}
			break;
		default:
			start += i;
			break;
		}
	}
	return start;
}

void *
oddjob_malloc(size_t size)
{
	void *ret;
	ret = malloc(size);
	if (ret == NULL) {
		fprintf(stderr, "Out of memory!\n");
		_exit(1);
	}
	return ret;
}

void *
oddjob_malloc0(size_t size)
{
	void *ret;
	ret = oddjob_malloc(size);
	memset(ret, 0, size);
	return ret;
}

void
oddjob_free(void *p)
{
	free(p);
}

char *
oddjob_strdup(const char *s)
{
	char *r;
	r = oddjob_malloc0(strlen(s) + 1);
	strcpy(r, s);
	return r;
}

char *
oddjob_strndup(const char *s, int n)
{
	char *r;
	int len;
	len = strlen(s);
	if (len > n) {
		len = n;
	}
	r = oddjob_malloc0(len + 1);
	memmove(r, s, len);
	return r;
}

void
oddjob_resize_array(void **array, size_t element_size,
		    size_t current_n_elements, size_t want_n_elements)
{
	void *p;
	if (want_n_elements > 0xffff) {
		fprintf(stderr, "Internal limit exceeded.\n");
		_exit(1);
	}
	if (element_size > 0xffff) {
		fprintf(stderr, "Internal limit exceeded.\n");
		_exit(1);
	}
	p = oddjob_malloc0(element_size * want_n_elements);
	if (current_n_elements > 0) {
		memmove(p, *array, element_size * current_n_elements);
		memset(*array, 0, element_size * current_n_elements);
		oddjob_free(*array);
	}
	*array = p;
}

char **
oddjob_collect_args(FILE *fp)
{
	char **ret, *thisline, *p;
	char buf[BUFSIZ];
	size_t l, m, arraysize;

	ret = NULL;
	thisline = NULL;
	arraysize = 0;
	while (fgets(buf, sizeof(buf), fp) != NULL) {
		char *tmp;
		l = thisline ? strlen(thisline) : 0;
		m = strlen(buf);
		tmp = oddjob_malloc0(l + m + 1);
		if (l > 0) {
			strcpy(tmp, thisline);
		}
		memmove(tmp + l, buf, m);
		oddjob_free(thisline);
		thisline = tmp;

		p = strpbrk(thisline, "\r\n");
		if (p != NULL) {
			*p = '\0';
			oddjob_resize_array((void**) &ret, sizeof(char*),
					    arraysize,
					    arraysize + 2);
			ret[arraysize] = thisline;
			arraysize++;
			thisline = NULL;
		}
	}
	if (thisline != NULL) {
		oddjob_resize_array((void**) &ret, sizeof(char*),
				    arraysize,
				    arraysize ?
				    arraysize + 1 : arraysize + 2);
		ret[arraysize] = thisline;
		arraysize = arraysize ? arraysize + 1 : arraysize + 2;
		thisline = NULL;
	}
	return ret;
}

void
oddjob_free_args(char **args)
{
	int i;
	if (args != NULL) {
		for (i = 0; args[i] != NULL; i++) {
			oddjob_free(args[i]);
			args[i] = NULL;
		}
		oddjob_free(args);
	}

}
